import React, { useLayoutEffect, useContext, useReducer, useRef, createContext as createContext$2, createElement, useState } from 'react';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var isBrowser = typeof window !== 'undefined' && typeof window.document !== 'undefined' && typeof window.document.createElement !== 'undefined' || process.env.NODE_ENV !== 'production' && process.env.TEST_MODE === 'Browser';

var useIsomorphicLayoutEffect = isBrowser ? useLayoutEffect : function (fn) {
  return fn();
};

var error = function error(message) {
  if (process.env.NODE_ENV === 'production') {
    return;
  }
  /* eslint no-console: 0 */


  console.error(message);
};
var log = {
  error: error
};

// https://github.com/reduxjs/react-redux (MIT LICENSE)
// https://github.com/dai-shi/use-context-selector (MIT LICENSE)

var createUseSelector = function createUseSelector(context, getState, subscription) {
  var defaultSelector = function defaultSelector(state) {
    return state;
  };

  function useSelector(selector) {
    var state = useContext(context);
    var storeSelector = selector || defaultSelector;

    var _useReducer = useReducer(function (s) {
      return s + 1;
    }, 0),
        _useReducer2 = _slicedToArray(_useReducer, 2),
        forceRender = _useReducer2[1];

    var selectedState;
    var latestSelector = useRef();
    var latestStoreState = useRef();
    var latestSelectedState = useRef();
    selectedState = latestSelectedState.current;

    try {
      if (storeSelector !== latestSelector.current || state !== latestStoreState.current) {
        selectedState = storeSelector(state);
      }
    } catch (err) {
      log.error(err);
    }

    useIsomorphicLayoutEffect(function () {
      latestSelector.current = storeSelector;
      latestStoreState.current = state;
      latestSelectedState.current = selectedState;
    });
    useIsomorphicLayoutEffect(function () {
      var refresh = function refresh() {
        var nextStore = getState();

        if (latestSelector.current === undefined || latestSelectedState.current === undefined) {
          return;
        }

        try {
          var newSelectedState = latestSelector.current(nextStore);

          if (newSelectedState === latestSelectedState.current) {
            return;
          }

          latestSelectedState.current = newSelectedState;
        } catch (err) {
          log.error(err);
        }

        forceRender();
      };

      subscription.add(refresh);
      return function () {
        subscription["delete"](refresh);
      };
    }, [context, subscription]);
    return selectedState;
  }

  return useSelector;
};

var createStore = function createStore(stateContext, dispatchContext, subscription, getCurrentState) {
  var useGlobalState = createUseSelector(stateContext, getCurrentState, subscription);

  var useGlobalDispatch = function useGlobalDispatch() {
    return useContext(dispatchContext);
  };

  return {
    useGlobalState: useGlobalState,
    useGlobalDispatch: useGlobalDispatch
  };
};

var dispatchEventLister = function dispatchEventLister(provider, subscription) {
  var dispatcher = /*#__PURE__*/React.memo(function (_ref) {
    var value = _ref.value,
        children = _ref.children;
    useIsomorphicLayoutEffect(function () {
      subscription.forEach(function (listener) {
        listener();
      });
    });
    return /*#__PURE__*/createElement(provider, {
      value: value
    }, children);
  });
  return dispatcher;
};

var createBaseContext = function createBaseContext(contextSource) {
  var stateContext = /*#__PURE__*/createContext$2(null, function () {
    return 0;
  });
  var dispatchContext = /*#__PURE__*/createContext$2(null);
  var subscription = new Set();
  stateContext.Provider = dispatchEventLister(stateContext.Provider, subscription);
  stateContext.displayName = 'UseGlobalStateContext';
  return {
    stateContext: stateContext,
    dispatchContext: dispatchContext,
    subscription: subscription
  };
};

var createCurrentState = function createCurrentState(initialState) {
  var currentState = _objectSpread2({}, initialState);

  var getCurrentState = function getCurrentState() {
    return currentState;
  };

  var setCurrentState = function setCurrentState(value, key) {
    currentState[key] = value;
  };

  var resetCurrentState = function resetCurrentState() {
    currentState = _objectSpread2({}, initialState);
  };

  return {
    getCurrentState: getCurrentState,
    setCurrentState: setCurrentState,
    resetCurrentState: resetCurrentState
  };
};

var entries = function entries(values) {
  return Object.entries(values);
};

var isFunction = function isFunction(value) {
  return value && {}.toString.call(value) === '[object Function]';
};

var createUseServerSideDispatch = function createUseServerSideDispatch(stateRef, getCurrentState, setCurrentState, displayName, subscription) {
  /* eslint no-param-reassign: 0 */
  function useServerSideDispatch(state) {
    var currentState = getCurrentState()[displayName];
    var newState = isFunction(state) ? state(currentState) : state;
    stateRef.current[displayName] = newState;
    setCurrentState(newState, displayName);
    subscription.forEach(function (listener) {
      listener();
    });
  }

  return useServerSideDispatch;
};

var createContext = function createContext(contextSource) {
  var _createCurrentState = createCurrentState(contextSource),
      getCurrentState = _createCurrentState.getCurrentState,
      setCurrentState = _createCurrentState.setCurrentState;

  var _createBaseContext = createBaseContext(),
      stateContext = _createBaseContext.stateContext,
      dispatchContext = _createBaseContext.dispatchContext,
      subscription = _createBaseContext.subscription;

  var _createStore = createStore(stateContext, dispatchContext, subscription, getCurrentState),
      useGlobalState = _createStore.useGlobalState,
      useGlobalDispatch = _createStore.useGlobalDispatch;

  var StateProvider = stateContext.Provider;
  var DispatchProvider = dispatchContext.Provider;

  var contextProvider = function contextProvider(_ref) {
    var children = _ref.children,
        value = _ref.value;
    var stateRef = useRef({});
    var dispatchRef = useRef({});
    entries(contextSource).forEach(function (_ref2) {
      var _ref3 = _slicedToArray(_ref2, 2),
          displayName = _ref3[0],
          initialState = _ref3[1];

      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;

      var _useState = useState(initialValue),
          _useState2 = _slicedToArray(_useState, 2),
          hookState = _useState2[0],
          hookDispatch = _useState2[1];

      setCurrentState(hookState, displayName);
      stateRef.current[displayName] = hookState;
      dispatchRef.current[displayName] = hookDispatch;
    }, {});
    return /*#__PURE__*/React.createElement(StateProvider, {
      value: stateRef.current
    }, /*#__PURE__*/React.createElement(DispatchProvider, {
      value: dispatchRef.current
    }, children));
  };

  var contextServerSideProvider = function contextServerSideProvider(_ref4) {
    var children = _ref4.children,
        value = _ref4.value;
    var stateRef = useRef({});
    var dispatchRef = useRef({});
    entries(contextSource).forEach(function (_ref5) {
      var _ref6 = _slicedToArray(_ref5, 2),
          displayName = _ref6[0],
          initialState = _ref6[1];

      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;
      useIsomorphicLayoutEffect(function () {
        setCurrentState(initialValue, displayName);
      }, []);
      var hookDispatch = createUseServerSideDispatch(stateRef, getCurrentState, setCurrentState, displayName, subscription);
      stateRef.current[displayName] = initialValue;
      dispatchRef.current[displayName] = hookDispatch;
    });
    return /*#__PURE__*/React.createElement(StateProvider, {
      value: stateRef.current
    }, /*#__PURE__*/React.createElement(DispatchProvider, {
      value: dispatchRef.current
    }, children));
  };

  return {
    useGlobalState: useGlobalState,
    useGlobalDispatch: useGlobalDispatch,
    contextProvider: isBrowser ? contextProvider : contextServerSideProvider,
    getState: getCurrentState
  };
};

/**
 * *useState* to create multiple contexts.
 * The created contexts are split into a state and a dispatch,
 * respectively, to prevent unnecessary rendering.
 */
var createUseStateContext = function createUseStateContext(contextSource) {
  var _createContext = createContext(contextSource),
      useGlobalState = _createContext.useGlobalState,
      useGlobalDispatch = _createContext.useGlobalDispatch,
      contextProvider = _createContext.contextProvider,
      getState = _createContext.getState;

  return [useGlobalState, useGlobalDispatch, contextProvider, getState];
};

var createStore$1 = function createStore(stateContext, dispatchContext, subscription, getCurrentState) {
  var useGlobalState = createUseSelector(stateContext, getCurrentState, subscription);

  var useGlobalDispatch = function useGlobalDispatch() {
    return useContext(dispatchContext);
  };

  return {
    useGlobalState: useGlobalState,
    useGlobalDispatch: useGlobalDispatch
  };
};

var getInitialState = function getInitialState(contextSource) {
  return entries(contextSource).reduce(function (acc, _ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        displayName = _ref2[0],
        initialState = _ref2[1].initialState;

    acc[displayName] = initialState;
    return acc;
  }, {});
};

var createUseServerSideDispatch$1 = function createUseServerSideDispatch(stateRef, getCurrentState, setCurrentState, displayName, reducer, subscription) {
  var useServerSideDispatch = function useServerSideDispatch(action) {
    /* eslint no-param-reassign: 0 */
    var currentState = getCurrentState()[displayName];
    var newState = reducer.length === 1 ? reducer(currentState, undefined) : reducer(currentState, action);
    stateRef.current[displayName] = newState;
    setCurrentState(newState, displayName);
    subscription.forEach(function (listener) {
      listener();
    });
  };

  return useServerSideDispatch;
};

var createContext$1 = function createContext(contextSource) {
  var _createCurrentState = createCurrentState(getInitialState(contextSource)),
      getCurrentState = _createCurrentState.getCurrentState,
      setCurrentState = _createCurrentState.setCurrentState;

  var _createBaseContext = createBaseContext(),
      stateContext = _createBaseContext.stateContext,
      dispatchContext = _createBaseContext.dispatchContext,
      subscription = _createBaseContext.subscription;

  var _createStore = createStore$1(stateContext, dispatchContext, subscription, getCurrentState),
      useGlobalState = _createStore.useGlobalState,
      useGlobalDispatch = _createStore.useGlobalDispatch;

  var StateProvider = stateContext.Provider;
  var DispatchProvider = dispatchContext.Provider;

  var contextProvider = function contextProvider(_ref3) {
    var children = _ref3.children,
        value = _ref3.value;
    var stateRef = useRef({});
    var dispatchRef = useRef({});
    entries(contextSource).forEach(function (_ref4) {
      var _ref5 = _slicedToArray(_ref4, 2),
          displayName = _ref5[0],
          source = _ref5[1];

      var reducer = source.reducer,
          initialState = source.initialState,
          initializer = source.initializer;
      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;

      var _useReducer = useReducer(reducer, initialValue, initializer),
          _useReducer2 = _slicedToArray(_useReducer, 2),
          hookState = _useReducer2[0],
          hookDispatch = _useReducer2[1];

      setCurrentState(hookState, displayName);
      stateRef.current[displayName] = hookState;
      dispatchRef.current[displayName] = hookDispatch;
    }, {});
    return /*#__PURE__*/React.createElement(StateProvider, {
      value: stateRef.current
    }, /*#__PURE__*/React.createElement(DispatchProvider, {
      value: dispatchRef.current
    }, children));
  };

  var contextServerSideProvider = function contextServerSideProvider(_ref6) {
    var children = _ref6.children,
        value = _ref6.value;
    var stateRef = useRef({});
    var dispatchRef = useRef({});
    entries(contextSource).forEach(function (_ref7) {
      var _ref8 = _slicedToArray(_ref7, 2),
          displayName = _ref8[0],
          initialState = _ref8[1].initialState;

      var initialValue = value && value[displayName] !== undefined ? value[displayName] : initialState;
      var hookDispatch = createUseServerSideDispatch$1(stateRef, getCurrentState, setCurrentState, displayName, contextSource[displayName].reducer, subscription);
      setCurrentState(initialValue, displayName);
      stateRef.current[displayName] = initialValue;
      dispatchRef.current[displayName] = hookDispatch;
    }, {});
    return /*#__PURE__*/React.createElement(StateProvider, {
      value: stateRef.current
    }, /*#__PURE__*/React.createElement(DispatchProvider, {
      value: dispatchRef.current
    }, children));
  };

  return {
    useGlobalState: useGlobalState,
    useGlobalDispatch: useGlobalDispatch,
    contextProvider: isBrowser ? contextProvider : contextServerSideProvider,
    getState: getCurrentState
  };
};

/**
 * *useReducer* to create multiple contexts.
 * The created contexts are split into a state and a dispatch,
 * respectively, to prevent unnecessary rendering.
 */
var createUseReducerContext = function createUseReducerContext(contextSource) {
  var _createContext = createContext$1(contextSource),
      useGlobalState = _createContext.useGlobalState,
      useGlobalDispatch = _createContext.useGlobalDispatch,
      contextProvider = _createContext.contextProvider,
      getState = _createContext.getState;

  return [useGlobalState, useGlobalDispatch, contextProvider, getState];
};

export { createUseReducerContext, createUseStateContext };
